import abc
from typing import Tuple

import numpy as np
from numpy.random import default_rng

from badgers.core.base import GeneratorMixin


class SeasonsGenerator(GeneratorMixin):
    """
    Base class for transformers that generate seasons in time-series data
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: a random number generator
        """
        self.random_generator = random_generator

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        pass


class GlobalAdditiveSinusoidalSeasonGenerator(SeasonsGenerator):
    """
    Add a sinusoidal season to the input time-series data
    """

    def __init__(self, random_generator=default_rng(seed=0), period: int = 10):
        super().__init__(random_generator=random_generator)
        self.period = period

    def generate(self, X, y, **params) -> Tuple:
        """

        :param X:
        :param y:
        :param params:
        :return:
        """
        t = np.arange(len(X))
        season = np.sin(t*2*np.pi/self.period)
        Xt = X + season
        return Xt, y



import unittest

from numpy.random import default_rng
from sklearn.datasets import make_blobs

from badgers.core.pipeline import Pipeline
from badgers.generators.tabular_data.imbalance import RandomSamplingClassesGenerator
from badgers.generators.tabular_data.noise import GaussianNoiseGenerator


class MyTestCase(unittest.TestCase):
    def test_generate(self):
        random_generator = default_rng(0)
        X, y = make_blobs(centers=3, random_state=0)
        generators = {
            'imbalance': RandomSamplingClassesGenerator(
                random_generator=random_generator,
                proportion_classes={0: 0.5, 1: 0.25, 2: 0.25}),
            'noise': GaussianNoiseGenerator(
                random_generator=random_generator,
                noise_std=0.5)
        }
        pipeline = Pipeline(generators=generators)
        Xt, yt = pipeline.generate(X, y)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import TestCase

import numpy as np
from numpy.random import default_rng

from badgers.generators.tabular_data.missingness import MissingValueGenerator
from tests.generators.tabular_data import generate_test_data_only_features


class TestMissingValueGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.generators_classes = MissingValueGenerator.__subclasses__()
        self.input_test_data = generate_test_data_only_features(rng=self.rng)

    def test_all_generators(self):
        for cls in self.generators_classes:
            transformer = cls()
            for input_type, (X, y) in self.input_test_data.items():
                with self.subTest(transformer=transformer.__class__, input_type=input_type):
                    Xt, _ = transformer.generate(X.copy(), y)
                    # assert arrays have same size
                    self.assertEqual(X.shape, Xt.shape)
                    # assert that the right number of nans have been generated
                    self.assertEqual(np.sum(np.isnan(Xt).sum()),
                                     transformer.percentage_missing / 100. * X.shape[0] * X.shape[1])


if __name__ == '__main__':
    unittest.main()



import functools
from enum import Enum

import pandas as pd


class TabularDataType(Enum):
    NUMPY_ARRAY = 1
    PANDAS_DATAFRAME = 2


def numpy_API(generate_func):
    """
    Ensures X can be worked on using the numpy API (useful for indexing!).
    If X is an object that does not strictly follow the numpy API (like pandas.DataFrame),
    then it internally stores the metadata (like columns), casts X to a numpy array, calls the generate function,
    and finally restore and restores the original type.

    @TODO check y too!
    """

    @functools.wraps(generate_func)
    def wrapper(self, X, y, **params):
        X_data_type = None
        if isinstance(X, pd.DataFrame):
            # when X is a pandas DataFrame, then locally save the columns and make X a numpy array
            X_data_type = TabularDataType.PANDAS_DATAFRAME
            columns = X.columns
            X = X.to_numpy()
        # call to generate function
        Xt, yt = generate_func(self, X, y, **params)
        if X_data_type == TabularDataType.PANDAS_DATAFRAME:
            Xt = pd.DataFrame(Xt, columns=columns)
        return Xt, yt

    return wrapper



import abc
from typing import Tuple

from numpy.random import default_rng
from sklearn.preprocessing import StandardScaler

from badgers.core.base import GeneratorMixin


class NoiseGenerator(GeneratorMixin):
    """
    Base class for transformers that add noise to tabular data
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: A random generator
        """
        self.random_generator = random_generator

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        pass


class LocalGaussianNoiseGenerator(NoiseGenerator):

    def __init__(self, random_generator=default_rng(seed=0), n_patterns: int = 10, patterns_width: int = 10,
                 noise_std: float = 0.1):
        super().__init__(random_generator=random_generator)
        self.n_patterns = n_patterns
        self.patterns_width = patterns_width
        self.noise_std = noise_std

    def generate(self, X, y, **params) -> Tuple:
        # TODO input validation!
        if X.ndim < 2:
            raise ValueError(
                "Expected 2D array. "
                "Reshape your data either using array.reshape(-1, 1) if "
                "your data has a single feature or array.reshape(1, -1) "
                "if it contains a single sample."
            )
        # generate extreme values indices and values
        self.patterns_indices_ = [(x, x + self.patterns_width) for x in
                                  self.random_generator.choice(X.shape[0] - self.patterns_width,
                                                               size=self.n_patterns,
                                                               replace=False, p=None)]

        scaler = StandardScaler()
        # fit, transform
        scaler.fit(X)
        Xt = scaler.transform(X)

        for (start, end) in self.patterns_indices_:
            Xt[start:end, :] += self.random_generator.normal(loc=0, scale=self.noise_std, size=(self.patterns_width, Xt.shape[1]))

        # inverse standardization
        return scaler.inverse_transform(Xt), y


class GlobalGaussianNoiseGenerator(NoiseGenerator):
    def __init__(self, random_generator=default_rng(seed=0), noise_std: float = 0.1):
        """

        :param random_generator: A random generator
        :param noise_std: The standard deviation of the noise to be added
        """
        super().__init__(random_generator=random_generator)
        self.noise_std = noise_std

    def generate(self, X, y, **params):
        """
        Add Gaussian white noise to the data.
        the data is first standardized (each column has a mean = 0 and variance = 1).
        The noise is generated from a normal distribution with standard deviation = `noise_std`.
        The noise is added to the data.

        :param X:
        :return:
        """
        scaler = StandardScaler()
        # fit, transform
        scaler.fit(X)
        Xt = scaler.transform(X)
        # add noise
        Xt = Xt + self.random_generator.normal(loc=0, scale=self.noise_std, size=Xt.shape)
        # inverse standardization
        return scaler.inverse_transform(Xt), y



from unittest import TestCase

import numpy as np

from badgers.generators.time_series.patterns import RandomConstantPatterns, RandomLinearPatterns


class TestRandomConstantPatterns(TestCase):

    def setUp(self) -> None:
        t = np.linspace(1, 10, 101)
        self.X = (np.sin(t * 2 * np.pi) + 0.5).reshape(-1, 1)

    def test_generate(self):
        generator = RandomConstantPatterns(n_patterns=3, patterns_width=5, constant_value=0)
        Xt, _ = generator.generate(self.X, None)
        for start, end in generator.patterns_indices_:
            self.assertListEqual(Xt[start:end, :].tolist(), np.zeros((end - start, self.X.shape[1])).tolist())


class TestRandomLinearPatterns(TestCase):

    def setUp(self) -> None:
        t = np.linspace(1, 10, 101)
        self.X = (np.sin(t * 2 * np.pi) + 0.5).reshape(-1, 1)

    def test_generate(self):
        generator = RandomLinearPatterns(n_patterns=3, patterns_width=5)
        Xt, _ = None, None
        for (start, end) in generator.patterns_indices_:
            for col in range(self.X.shape[1]):
                self.assertListEqual(Xt[start:end, col].tolist(),
                                     np.linspace(self.X[start, col], self.X[end, col], end - start).tolist())



import abc
from typing import Tuple

import numpy as np
from numpy.random import default_rng

from badgers.core.base import GeneratorMixin


class TrendsGenerator(GeneratorMixin):
    """
    Base class for transformers that generate trends in time-series data
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: a random number generator
        """
        self.random_generator = random_generator

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        pass


class GlobalAdditiveLinearTrendGenerator(TrendsGenerator):
    """
    Add a linear trend to the input time-series data
    """

    def __init__(self, random_generator=default_rng(seed=0), slope: float = 0.1):
        super().__init__(random_generator=random_generator)
        self.slope = slope

    def generate(self, X, y, **params) -> Tuple:
        """

        :param X:
        :param y:
        :param params:
        :return:
        """
        trend = self.slope * np.linspace(0, 1, len(X))
        Xt = X + trend
        return Xt, y



from unittest import TestCase

from numpy.random import default_rng

from badgers.generators.tabular_data.drift import RandomShiftGenerator, RandomShiftClassesGenerator
from tests.generators.tabular_data import generate_test_data_with_classification_labels, \
    generate_test_data_only_features


class TestRandomShiftGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.test_data = generate_test_data_only_features(rng=self.rng)

    def test_generate(self):
        """
        Asserts:
        - Arrays X and Xt have the same shape
        """
        generator = RandomShiftGenerator(random_generator=self.rng)
        for input_type, (X, y) in self.test_data.items():
            with self.subTest(transformer=generator.__class__, input_type=input_type):
                Xt, _ = generator.generate(X.copy(), y)
                # assert arrays have same size
                self.assertEqual(X.shape, Xt.shape)


class TestRandomShiftGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.test_data = generate_test_data_with_classification_labels(rng=self.rng)

    def test_generate(self):
        """
        Asserts:
        - Arrays X and Xt have the same shape
        """
        generator = RandomShiftClassesGenerator(random_generator=self.rng, shift_std=0.1)
        for input_type, (X, y) in self.test_data.items():
            with self.subTest(transformer=generator.__class__, input_type=input_type):
                Xt, yt = generator.generate(X.copy(), y)
                # assert arrays have same size
                self.assertEqual(X.shape, Xt.shape)



import unittest
from numpy.random import default_rng

from badgers.generators.time_series.changepoints import ChangePointGenerator, RandomChangeInMeanGenerator


class TestChangePointGenerator(unittest.TestCase):
    def setUp(self):
        self.random_generator = default_rng(seed=0)
        self.n_changepoints = 10
        self.min_change = -5
        self.max_change = 5

    def test_ChangePointGenerator_init(self):
        generator = ChangePointGenerator(random_generator=self.random_generator, n_changepoints=self.n_changepoints)
        self.assertEqual(generator.random_generator, self.random_generator)
        self.assertEqual(generator.n_changepoints, self.n_changepoints)
        self.assertEqual(generator.changepoints, None)

    def test_RandomChangeInMeanGenerator_init(self):
        generator = RandomChangeInMeanGenerator(random_generator=self.random_generator, n_changepoints=self.n_changepoints, min_change=self.min_change, max_change=self.max_change)
        self.assertEqual(generator.random_generator, self.random_generator)
        self.assertEqual(generator.n_changepoints, self.n_changepoints)
        self.assertEqual(generator.min_change, self.min_change)
        self.assertEqual(generator.max_change, self.max_change)
        self.assertEqual(generator.changepoints, None)

    def test_RandomChangeInMeanGenerator_generate(self):
        cp_generator = RandomChangeInMeanGenerator(random_generator=self.random_generator, n_changepoints=self.n_changepoints, min_change=self.min_change, max_change=self.max_change)
        X = [1, 2, 3, 4, 5]
        y = None

        Xt, _ = cp_generator.generate(X, y)
        self.assertNotEqual(Xt, X)
        self.assertEqual(y, y)


if __name__ == '__main__':
    unittest.main()


import abc
from typing import Tuple

import numpy as np
from numpy.random import default_rng

from badgers.core.base import GeneratorMixin


class PatternsGenerator(GeneratorMixin):
    """
    Base class for transformers that generate patterns in time-series data
    """

    def __init__(self, random_generator=default_rng(seed=0), n_patterns: int = 10):
        """
        :param random_generator: a random number generator
        :param n_patterns: the number of patterns to generate
        """
        self.random_generator = random_generator
        self.n_patterns = n_patterns
        self.patterns_indices_ = []

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        pass


class RandomConstantPatterns(PatternsGenerator):
    """
    Randomly generates constant patterns of constant value
    """

    def __init__(self, random_generator=default_rng(seed=0), n_patterns: int = 10, patterns_width: int = 10,
                 constant_value: float = 0):
        super().__init__(random_generator=random_generator, n_patterns=n_patterns)
        self.patterns_width = patterns_width
        self.constant_value = constant_value

    def generate(self, X, y, **params) -> Tuple:
        # TODO input validation!
        if X.ndim < 2:
            raise ValueError(
                "Expected 2D array. "
                "Reshape your data either using array.reshape(-1, 1) if "
                "your data has a single feature or array.reshape(1, -1) "
                "if it contains a single sample."
            )
        # generate extreme values indices and values
        self.patterns_indices_ = [(x, x + self.patterns_width) for x in
                                  self.random_generator.choice(X.shape[0] - self.patterns_width, size=self.n_patterns,
                                                               replace=False, p=None)]

        for (start, end) in self.patterns_indices_:
            X[start:end, :] = self.constant_value

        return X, y


class RandomLinearPatterns(PatternsGenerator):
    """
        Randomly generates patterns of constant slope (linear)
        """

    def __init__(self, random_generator=default_rng(seed=0), n_patterns: int = 10, patterns_width: int = 10):
        super().__init__(random_generator=random_generator, n_patterns=n_patterns)
        self.patterns_width = patterns_width

    def generate(self, X, y, **params) -> Tuple:
        # TODO input validation!
        if X.ndim < 2:
            raise ValueError(
                "Expected 2D array. "
                "Reshape your data either using array.reshape(-1, 1) if "
                "your data has a single feature or array.reshape(1, -1) "
                "if it contains a single sample."
            )
        # generate extreme values indices and values
        self.patterns_indices_ = [(x, x + self.patterns_width) for x in
                                  self.random_generator.choice(X.shape[0] - self.patterns_width, size=self.n_patterns,
                                                               replace=False, p=None)]

        for (start, end) in self.patterns_indices_:
            for col in range(X.shape[1]):
                X[start:end, col] = np.linspace(X[start,col],X[end,col], end - start)

        return X, y



import abc

import numpy as np
from numpy.random import default_rng
from sklearn.preprocessing import StandardScaler

from badgers.core.base import GeneratorMixin


class NoiseGenerator(GeneratorMixin):
    """
    Base class for generators that add noise to tabular data
    """

    def __init__(self, random_generator=default_rng(seed=0), repeat=1):
        """
        :param random_generator: A random generator
        :param repeat: number of times a noisy point is generated from the original.
            repeat = 1 means that Xt.shape[0] == X.shape[0], repeat = 10 means that Xt.shape[0] == 10 * X.shape[0]
        """
        self.random_generator = random_generator
        self.repeat = repeat

    @abc.abstractmethod
    def generate(self, X, y=None, **params):
        pass


class GaussianNoiseGenerator(NoiseGenerator):
    """
    A generator that adds Gaussian white noise to the tabular data
    """

    def __init__(self, random_generator=default_rng(seed=0), noise_std: float = 0.1, repeat=1):
        """

        :param random_generator: A random generator
        :param noise_std: The standard deviation of the noise to be added
        """
        super().__init__(random_generator=random_generator, repeat=repeat)
        self.noise_std = noise_std

    def generate(self, X, y, **params):
        """
        Adds Gaussian white noise to the data.
        The data is first standardized (each column has a mean = 0 and variance = 1).
        The noise is generated from a normal distribution with standard deviation = `noise_std`.
        The noise is added to the data.

        :param X: the input
        :param y: the target
        :param params: optional parameters
        :return: Xt, yt
        """
        # standardize X
        scaler = StandardScaler()
        # fit, transform
        scaler.fit(X)
        Xt = scaler.transform(X)
        # add noise and repeat
        Xt = np.concatenate(
            [
                Xt + self.random_generator.normal(loc=0, scale=self.noise_std, size=Xt.shape)
                for _ in range(self.repeat)
            ], axis=0
        )
        if y is not None:
            yt = np.concatenate([y] * self.repeat, axis=0)
        else:
            yt = None
        # inverse pca
        return scaler.inverse_transform(Xt), yt


class GaussianNoiseClassesGenerator(NoiseGenerator):
    """
    A generator that adds Gaussian white noise to each class separately.
    """

    def __init__(self, random_generator=default_rng(seed=0), repeat=1, noise_std_per_class: dict = None):
        """

        :param random_generator: A random generator
        :param noise_std_per_class: A dictionary giving the standard deviation of the noise to be added for each class
            key = class labels, values = noise std for this given class
        """
        super().__init__(random_generator=random_generator, repeat=repeat)
        self.noise_std_per_class = noise_std_per_class

    def generate(self, X, y, **params):
        """
        Add Gaussian white noise to the data.
        the data is first standardized (each column has a mean = 0 and variance = 1).
        The noise is generated from a normal distribution with standard deviation = `noise_std`.
        The noise is added to the data.

        :param X: the input
        :param y: the target
        :param params: optional parameters
        :return: Xt, yt
        """
        # standardize X
        scaler = StandardScaler()
        # fit, transform
        scaler.fit(X)
        Xt = scaler.transform(X)
        # add noise and repeat for each class

        tmp_Xt = []
        tmp_yt = []
        for label, noise_std in self.noise_std_per_class.items():
            data_class = np.array(Xt[y == label])
            noisy_data_class = np.concatenate(
                [
                    data_class + self.random_generator.normal(loc=0, scale=noise_std, size=data_class.shape)
                    for _ in range(self.repeat)
                ],
                axis=0
            )
            labels = [label] * self.repeat * data_class.shape[0]
            tmp_Xt.append(noisy_data_class.copy())
            tmp_yt.append(labels)

        Xt = np.concatenate(tmp_Xt, axis=0)
        yt = np.concatenate(tmp_yt, axis=0)
        # inverse pca
        return scaler.inverse_transform(Xt), yt



import unittest

import numpy as np
from numpy.random import default_rng

from badgers.generators.time_series.trends import GlobalAdditiveLinearTrendGenerator


class TestTrendsGenerator(unittest.TestCase):
    def setUp(self):
        self.random_generator = default_rng(seed=0)

    def test_global_additive_linear_trend_generator(self):
        # Test the generate method of GlobalAdditiveLinearTrendGenerator
        global_additive_linear_trend_generator = GlobalAdditiveLinearTrendGenerator(
            random_generator=self.random_generator, slope=1)
        X = np.zeros(100)
        y = None
        expected_Xt = np.linspace(0, 1, len(X))

        Xt, _ = global_additive_linear_trend_generator.generate(X, y)

        self.assertListEqual(list(Xt), list(expected_Xt))


if __name__ == '__main__':
    unittest.main()



import abc
from typing import Tuple

from numpy.random import default_rng

from badgers.core.base import GeneratorMixin


class ChangePointGenerator(GeneratorMixin):
    """
    Base class for generators that generate changepoints in time-series data
    """

    def __init__(self, random_generator=default_rng(seed=0), n_changepoints: int = 10):
        """
        :param random_generator: a random number generator
        :param n_outliers: the number of outliers to generate
        """
        self.random_generator = random_generator
        self.n_changepoints = n_changepoints
        self.changepoints = None

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        pass


class RandomChangeInMeanGenerator(ChangePointGenerator):
    """
    Generate randomly change in mean changepoints
    """

    def __init__(self, random_generator=default_rng(seed=0), n_changepoints: int = 10, min_change: float = -5,
                 max_change: float = 5):
        super().__init__(random_generator=random_generator, n_changepoints=n_changepoints)
        self.min_change = min_change
        self.max_change = max_change

    def generate(self, X, y, **params) -> Tuple:
        """

        :param X:
        :param y:
        :param params:
        :return:
        """
        # Generate change points
        self.changepoints = list(
            zip(
                self.random_generator.integers(int(0.05 * len(X)), int(0.95 * len(X)), size=self.n_changepoints),
                self.random_generator.uniform(self.min_change, self.max_change, size=self.n_changepoints)
            )
        )

        Xt = X.copy()

        for idx, change in self.changepoints:
            Xt[idx:] += change

        return Xt, y



import unittest
from unittest import TestCase

import numpy as np
from numpy.random import default_rng

from badgers.core.utils import normalize_proba
from badgers.generators.tabular_data.imbalance import RandomSamplingFeaturesGenerator, \
    RandomSamplingClassesGenerator, RandomSamplingTargetsGenerator
from tests.generators.tabular_data import generate_test_data_with_classification_labels, \
    generate_test_data_with_regression_targets


class TestRandomSamplingClassesGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.input_test_data = generate_test_data_with_classification_labels(rng=self.rng)

    def test_generate(self):

        for input_type, (X, y) in self.input_test_data.items():
            if input_type in ['numpy_1D', 'pandas_1D']:
                proportion_classes = {0: 0.9, 1: 0.1}
            elif input_type in ['numpy_2D', 'pandas_2D']:
                proportion_classes = {0: 0.5, 1: 0.2, 2: 0.1, 3: 0.1, 4: 0.1}
            else:
                self.fail()
            transformer = RandomSamplingClassesGenerator(proportion_classes=proportion_classes)

            Xt, yt = transformer.generate(X.copy(), y)
            # assert arrays have same size
            self.assertEqual(Xt.shape[1], X.shape[1])
            self.assertEqual(Xt.shape[0], len(yt))
            # assert pandas DataFrame get the same columns before and after the call to generate function
            if input_type in ['pandas_1D', 'pandas_2D']:
                self.assertListEqual(list(X.columns), list(Xt.columns))


class TestRandomSamplingFeaturesGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.input_test_data = generate_test_data_with_classification_labels(rng=self.rng)

    def test_generate(self):
        def proba_func(X):
            feature = X[:, 0]
            return normalize_proba(
                (np.max(feature) - feature) / (np.max(feature) - np.min(feature))
            )

        transformer = RandomSamplingFeaturesGenerator(sampling_proba_func=proba_func)

        for input_type, (X, y) in self.input_test_data.items():
            with self.subTest(transformer=transformer.__class__, input_type=input_type):
                Xt, _ = transformer.generate(X.copy(), y)
                # assert arrays have same size
                self.assertEqual(Xt.shape[1], X.shape[1])
                self.assertEqual(Xt.shape[0], X.shape[0])
                # assert pandas DataFrame get the same columns before and after the call to generate function
                if input_type in ['pandas_1D', 'pandas_2D']:
                    self.assertListEqual(list(X.columns), list(Xt.columns))


class TestRandomSamplingTargetsGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.input_test_data = generate_test_data_with_regression_targets(rng=self.rng)

    def test_generate(self):
        def proba_func(y):
            return normalize_proba(
                (np.max(y) - y) / (np.max(y) - np.min(y))
            )

        transformer = RandomSamplingTargetsGenerator(sampling_proba_func=proba_func)

        for input_type, (X, y) in self.input_test_data.items():
            with self.subTest(transformer=transformer.__class__, input_type=input_type):
                Xt, _ = transformer.generate(X.copy(), y)
                # assert arrays have same size
                self.assertEqual(Xt.shape[1], X.shape[1])
                self.assertEqual(Xt.shape[0], X.shape[0])
                # assert pandas DataFrame get the same columns before and after the call to generate function
                if input_type in ['pandas_1D', 'pandas_2D']:
                    self.assertListEqual(list(X.columns), list(Xt.columns))


if __name__ == '__main__':
    unittest.main()
